/////////////////////////////////////////////////////////////////////////////
// DVIDraw.cpp
// 
// Copyright (C) David McCabe, 2001. <sunlightd@hotmail.com>
//
//  This file is part of Embedded TeX.
//  
//	Embedded TeX is free software; you can redistribute it and/or 
//	modify it under the terms of the GNU General Public License as 
//	published by the Free Software Foundation; either version 2, 
//	or (at your option) any later version.
//		
//	Embedded TeX is distributed in the hope that it will be useful, 
//	but WITHOUT ANY WARRANTY; without even the implied warranty of 
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
//	See the GNU General Public License for more details.
//	
//	You should have received a copy of the GNU General Public 
//	License along with Embedded TeX; if not, write to the Free 
//	Software Foundation, 59 Temple Place - Suite 330, Boston, 
//	MA 02111-1307, USA.
//
//	NOTE: This module is based on prior work as noted below:
//  
//		Yap: Copyright (C) 2000 Christian Schenk <cs@miktex.org>
//
//	This file holds the DVI loading and simple drawing code.
//	Specials are implemented in tpic.cpp.

#include "StdAfx.h"
#include "DVIDraw.h"
#include "Globals.h"

/////////////////////////////////////////////////////////////////////////////
// CDVIDraw

/////////////////////////////////////////////////////////////////////////////
// CDVIDraw construction/destruction

CDVIDraw::CDVIDraw()
{
	m_dvi = NULL;
	m_pPage = NULL;

	// Initialize monochrome DIB for glyphs
	void *p = malloc(sizeof(BITMAPINFO) + 2 * (sizeof(RGBQUAD)));
	ATLASSERT(p != NULL);

	m_pbmi = (LPBITMAPINFO)p;
	m_pbmi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	m_pbmi->bmiHeader.biPlanes = 1;
	m_pbmi->bmiHeader.biBitCount = 1;
	m_pbmi->bmiHeader.biCompression = BI_RGB;
	m_pbmi->bmiHeader.biClrUsed = 0;
	m_pbmi->bmiHeader.biClrImportant = 0;
	m_pbmi->bmiHeader.biXPelsPerMeter = 
		m_pbmi->bmiHeader.biYPelsPerMeter = (GetResolution() * 10000) / 254;
	
	// Initialise colour table. Note that 0 is white, and 1 is black...
	RGBQUAD *pCol = m_pbmi->bmiColors;
	pCol[0].rgbRed = pCol[0].rgbGreen = pCol[0].rgbBlue = 255;
	pCol[0].rgbReserved = 0;
	pCol[1].rgbRed = pCol[1].rgbGreen = pCol[1].rgbBlue = pCol[1].rgbReserved = 0;
}

CDVIDraw::~CDVIDraw ()
{
	if (m_pbmi != NULL)
		free(m_pbmi);

	if (m_dvi)
		MiKTeX::dvi::destroy(m_dvi);
}

/////////////////////////////////////////////////////////////////////////////
// CDVIDraw drawing

void CDVIDraw::DrawPage(HDC hDC, int x, int y)
{
	// Calculate horizontal/vertical offsets
	RECT	rBounds;
	
	GetBoundingBox(&rBounds);

	// Set viewport origin for current page
	x += -rBounds.left;
	y += -rBounds.top;
	
	// Draw background rules
	DrawRules(hDC, true, x, y);
	
	// Draw text
	DrawGlyphs(hDC, x, y);
	
	// Draw foreground rules
	DrawRules(hDC, false, x, y);

}

// Draw text
void CDVIDraw::DrawGlyphs(HDC hDC, int x, int y)
{
	ATLASSERT(GetDeviceCaps(hDC, RASTERCAPS) & RC_STRETCHDIB);
	
	ATLASSERT(m_pPage != NULL);
	
	SetStretchBltMode(hDC, COLORONCOLOR);
	
	int				nBitmaps = m_pPage->numofbitmaps();
	LPBITMAPINFO	bmi;
		
	for (int i = 0; i < nBitmaps; i++)
	{
		const	MiKTeX::dvibitmap& dvibm = m_pPage->getbitmap(i);
		
		RECT rectBitmap;
		
		SetRect(&rectBitmap, dvibm.x + x, dvibm.y + y,
			dvibm.x + dvibm.width - 1 + x, dvibm.y + dvibm.height - 1 + y);
		
		bmi = MakeBitmapInfo(dvibm.width, dvibm.height, dvibm.bytesperline);
		
  		StretchDIBits(hDC,
			dvibm.x + x, dvibm.y + y, 
			dvibm.width, dvibm.height,
			0, 0, dvibm.width, dvibm.height,
			dvibm.bitmap, bmi,
			DIB_RGB_COLORS, SRCCOPY);
	}
}

// Construct a LPBITMAPINFO with the specified dimensions
LPBITMAPINFO CDVIDraw::MakeBitmapInfo(int width, int height, int bytesperline)
{
	m_pbmi->bmiHeader.biWidth = width;
	m_pbmi->bmiHeader.biHeight = height;
	ATLASSERT(bytesperline % 4 == 0);
	m_pbmi->bmiHeader.biSizeImage = bytesperline * height;
	return m_pbmi;
}

// Draw rules
void CDVIDraw::DrawRules(HDC hDC, bool bDrawBlackboards, int x, int y)
{
	ATLASSERT(m_pPage != NULL);
	
	const MiKTeX::dvirule	*pRule;
	
	int		i;
	HBRUSH	hBrush, hOldBrush;

	for (i = 0; (pRule = m_pPage->getrule(i)) != NULL; i++)
	{
		if (pRule->isblackboard() && !bDrawBlackboards)
			continue;
		
		if (!pRule->isblackboard() && bDrawBlackboards)
			continue;
		
		RECT rectRule;
		
		SetRect(&rectRule, pRule->left() + x, pRule->top() + y,
			(pRule->right() + 1) + x, (pRule->bottom() + 1) + y);

		if ((rectRule.right - rectRule.left) < 1)
			rectRule.right = rectRule.left + 1;
		if ((rectRule.bottom - rectRule.top) < 1)
			rectRule.bottom = rectRule.top + 1;

		hBrush = CreateSolidBrush(pRule->background());
		hOldBrush = (HBRUSH)SelectObject(hDC, hBrush);
		FillRect(hDC, &rectRule, hBrush);
		SelectObject(hDC, hOldBrush);
		DeleteObject(hBrush);
	}
}

/////////////////////////////////////////////////////////////////////////////
// CDVIDraw file loading

// Callback for loading with messages
bool ProgressCallback(const char *lpszMsg)
{
	return true;
}

// Load a DVI format file into this object.
BOOL CDVIDraw::OpenDVI(LPCTSTR lpszPathName) 
{
	ATLASSERT(m_dvi == NULL);
	
	MIKTEXMFMODE mfmode;
	
	if (!MiKTeX::GetMETAFONTMode(g_nMetafontMode, &mfmode))
		return FALSE;

	try
    {
		// Load the file.
		m_dvi = MiKTeX::dvi::create(lpszPathName, mfmode.szMnemonic, GetResolution(), 1);
		// Scan the file.
		m_dvi->scan();

		if (m_dvi->pagecount() == 0)
		{
			MiKTeX::dvi::destroy(m_dvi);
			m_dvi = NULL;
			return FALSE;
		}

		// Load the first page.
		m_dvi->loadpage(0, ProgressCallback);
		m_dvi->gotopage(0);

		m_pPage = m_dvi->getpage(0);
		if (m_pPage == NULL)
		{
			MiKTeX::dvi::destroy(m_dvi);
			m_dvi = NULL;
			return FALSE;
		}

		return TRUE;
    }
	catch(...)
	{
		if (m_dvi != NULL)
			MiKTeX::dvi::destroy(m_dvi);
		m_dvi = NULL;
		return FALSE;
	}
}

/////////////////////////////////////////////////////////////////////////////
// CDVIDraw commands

// Get the display resolution
int CDVIDraw::GetResolution() const
{
	MIKTEXMFMODE mfmode;
	
	MiKTeX::GetMETAFONTMode(g_nMetafontMode, &mfmode);
	return mfmode.iHorzRes;
}

// Find a rectangle enclosing all the objects on the page
void CDVIDraw::GetBoundingBox(LPRECT pRect)
{
	ATLASSERT(m_pPage != NULL);
	
	const MiKTeX::dvirule *pRule;
	RECT	r;
	int		i, nBitmaps = m_pPage->numofbitmaps();
	
	::SetRectEmpty(&r);

	// Search glyphs
	for (i = 0; i < nBitmaps; i++)
	{
		const MiKTeX::dvibitmap & dvibm = m_pPage->getbitmap(i);
		
		RECT rectBitmap;
		
		SetRect(&rectBitmap, dvibm.x, dvibm.y, dvibm.x + dvibm.width + 1, dvibm.y + dvibm.height + 1);

		UnionRect(&r, &r, &rectBitmap);
	}

	// Search rules
	for (i = 0; (pRule = m_pPage->getrule(i)) != 0; i++)
	{
		RECT rectRule;
		
		SetRect(&rectRule, pRule->left(), pRule->top(), pRule->right() + 1, pRule->bottom() + 1);

		UnionRect(&r, &r, &rectRule);
	}

	*pRect = r;
}
